﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Runtime.Serialization.Json;
using System.Xml;
using System.Xml.Linq;
using System.Xml.XPath;
using System.IO;

namespace asprise_imaging_api
{
    /// <summary>
    /// Represents a scan request.
    /// </summary>
    public class Request : ICloneable
    {

        /// <summary>
        /// Defaule blank page threshold
        /// </summary>
    public const decimal DEFAULT_BLANK_PAGE_THRESHOLD = 0.02m;

        /// <summary>
        /// Default blank page margin percent.
        /// </summary>
    public const int DEFAULT_BLANK_PAGE_MARGIN_PERCENT = 8;

    /** Request ID */
    String id;

    /**
     * By default images will be processed after each scan; Set to true to process images after all scans.
     * Processing images after all scan could speed up the scanning process.
     */
    Boolean processImagesAfterAllScans = false;

    /** Whether to prompt scan more dialog. */
    Boolean promptScanMore = false;

    /** Whether blank pages should be discard. */
    Boolean discardBlankPages = false;

    /** Ink coverage threshold to determine if a page is blank. */
    decimal blankPageThreshold = DEFAULT_BLANK_PAGE_THRESHOLD;

    /** Percentage of margin to be excluded when calculating ink coverage. */
    int blankPageMarginPercent = DEFAULT_BLANK_PAGE_MARGIN_PERCENT;

    /** Whether barcodes should be recognized. */
    Boolean recognizeBarcodes;

    /** At which DPI barcodes will be recognized. */
    int barcodesDpi;

    /** Additional barcode recognition settings */
    String barcodesSettings;

    /** Map of cap name (could be appended with operation and priority, e.g., /RESET, /SET/3456) */
    Dictionary<String, String> twainCapSetting = new Dictionary<String, String>();

    /** Names of caps to be retrieved. */
    List<String> retrieveCaps = new List<String>();

    /** Name of extended image attributes to be retrieved. */
    List<String> retrieveExtendedImageAttributes = new List<String>();

    /** Output specifications. */
    List<RequestOutputItem> outputItems = new List<RequestOutputItem>();

    /** List of image files. */
    List<string> imageFiles = new List<string>();

    Boolean useAspriseDialog = false;
    String dialogTitle = null;
    int dialogWidth = -1;
    int dialogHeight = -1;
    Boolean showScannerUI = false;
    Boolean modalScannerUI = false;
    String sourceName = null;
    Boolean onWebServer = false;

    /** The root of JSON object that this request is deserialized from. */
    //Map<Object, Object> sourceJson;

    /// <summary>
    /// Deep clones.
    /// </summary>
    /// <returns></returns>
    public Object Clone() {
        Request clone = (Request) this.MemberwiseClone();
        clone.twainCapSetting = clone.twainCapSetting == null ? null : new Dictionary<string, string>(clone.twainCapSetting);
        clone.retrieveCaps = clone.retrieveCaps == null ? null : new List<string>(clone.retrieveCaps);
        clone.retrieveExtendedImageAttributes = 
            clone.retrieveExtendedImageAttributes == null ? null : new List<string>(clone.retrieveExtendedImageAttributes);
        clone.outputItems = new List<RequestOutputItem>();
        foreach(RequestOutputItem item in outputItems) {
            clone.outputItems.Add((RequestOutputItem) item.Clone());
        }
        clone.imageFiles = new List<string>(imageFiles);
        return clone;
    }

    /**
     * Constructor.
     */
    public Request() {
        id = ((DateTime.UtcNow.Subtract(new DateTime(1970, 1, 1))).TotalMilliseconds).ToString();
    }

    /// <summary>
    /// Deserialize from JSON strings.
    /// </summary>
    /// <param name="spec">The input json string</param>
    /// <returns></returns>
    public static Request FromJson(String spec) {
        XElement root = ScanUtils.parseJson(spec);

        Request request = new Request();
        request.id = ScanUtils.xeElementValueAsString(root, "request_id", null);
        request.processImagesAfterAllScans = 
            ScanUtils.xeElementValueAsString(root, "processing_strategy", "default").Equals("after-all-scans");
        request.promptScanMore = ScanUtils.xeElementValueAsString(root, "prompt_scan_more", ScanUtils.STR_false).Equals(ScanUtils.STR_true);
        request.discardBlankPages = ScanUtils.xeElementValueAsString(root, "discard_blank_pages", ScanUtils.STR_false).Equals(ScanUtils.STR_true);
        request.blankPageThreshold = ScanUtils.xeElementValueAsDecimal(root, "blank_page_threshold", DEFAULT_BLANK_PAGE_THRESHOLD);
        request.blankPageMarginPercent = ScanUtils.xeElementValueAsInt(root, "blank_page_margin_percent", DEFAULT_BLANK_PAGE_MARGIN_PERCENT);
        request.recognizeBarcodes = ScanUtils.xeElementValueAsString(root, "recognize_barcodes", ScanUtils.STR_false).Equals(ScanUtils.STR_true);
        request.barcodesDpi = ScanUtils.xeElementValueAsInt(root, "barcodes_dpi", -1);
        request.barcodesSettings = ScanUtils.xeElementValueAsString(root, "barcodes_settings", null);

        request.useAspriseDialog = ScanUtils.xeElementValueAsString(root, "use_asprise_dialog", ScanUtils.STR_false).Equals(ScanUtils.STR_true);
        request.showScannerUI = ScanUtils.xeElementValueAsString(root, "show_scanner_ui", ScanUtils.STR_false).Equals(ScanUtils.STR_true);
        request.modalScannerUI = ScanUtils.xeElementValueAsString(root, "modal_scanner_ui", ScanUtils.STR_true).Equals(ScanUtils.STR_true);
        request.dialogWidth = ScanUtils.xeElementValueAsInt(root, "dialog_width", -1);
        request.dialogHeight = ScanUtils.xeElementValueAsInt(root, "dialog_height", -1);
        request.sourceName = ScanUtils.xeElementValueAsString(root, "source_name", null);

        request.onWebServer = ScanUtils.xeElementValueAsString(root, "on_web_server", ScanUtils.STR_false).Equals(ScanUtils.STR_true);

        request.twainCapSetting = ScanUtils.toDictStringString(ScanUtils.xeElementObjectToDictionary(root.Element("twain_cap_setting")));

        XElement xeOutputSettings = root.Element("output_settings");
        if(xeOutputSettings != null) {
            foreach(XElement xeOutputSetting in xeOutputSettings.Elements()) {
                request.outputItems.Add(RequestOutputItem.FromJsonMap(xeOutputSetting));
            }
        }

        request.retrieveCaps = ScanUtils.toListOfString(ScanUtils.xeElementArrayToList(root.Element("retrieve_caps")));

        request.retrieveExtendedImageAttributes = ScanUtils.toListOfString(ScanUtils.xeElementArrayToList(root.Element("retrieve_extended_image_attrs")));

        request.imageFiles = ScanUtils.toListOfString(ScanUtils.xeElementArrayToList(root.Element("image_files")));

        return request;
    }

        /// <summary>
        /// JSON serialization.
        /// </summary>
        /// <param name="pretty"></param>
        /// <returns></returns>
    public String ToJson(Boolean pretty = false)
    {
        using (MemoryStream ms = new MemoryStream())
        {
            XmlDictionaryWriter writer = JsonReaderWriterFactory.CreateJsonWriter(ms);
            writer.WriteStartDocument(true);
            ToJsonObject(writer);
            writer.WriteEndDocument();
            writer.Flush();

            ms.Flush();
            ms.Position = 0;
            return new StreamReader(ms).ReadToEnd();
        }
    }

        /// <summary>
        /// Serializes to JSON.
        /// </summary>
        /// <param name="writer"></param>
    public void ToJsonObject(XmlDictionaryWriter writer) {
        writer.WriteStartElement("root");
        writer.WriteAttributeString("type", "object");

        writer.WriteElementString("request_id", id);
        if (processImagesAfterAllScans)
        {
            writer.WriteElementString("processing_strategy", "after-all-scans");
        }

        if (twainCapSetting != null)
        {
            ScanUtils.writeElementDict(writer, "twain_cap_setting", twainCapSetting);
        }

        writer.WriteElementString("prompt_scan_more", promptScanMore ? ScanUtils.STR_true : ScanUtils.STR_false);
        if (discardBlankPages)
        {
            writer.WriteElementString("discard_blank_pages", discardBlankPages ? ScanUtils.STR_true : ScanUtils.STR_false);
            ScanUtils.writeElementDecimal(writer, "blank_page_threshold", blankPageThreshold);
            ScanUtils.writeElementInt(writer, "blank_page_margin_percent", blankPageMarginPercent);
        }

        if (recognizeBarcodes)
        {
            writer.WriteElementString("recognize_barcodes", recognizeBarcodes ? ScanUtils.STR_true : ScanUtils.STR_false);
        }
        if (barcodesDpi > 0)
        {
            ScanUtils.writeElementInt(writer, "barcodes_dpi", barcodesDpi);
        }
        if (barcodesSettings != null && barcodesSettings.Trim().Length > 0)
        {
            writer.WriteElementString("barcodes_settings", barcodesSettings);
        }

        if(onWebServer) {
            writer.WriteElementString("on_web_server", onWebServer ? ScanUtils.STR_true : ScanUtils.STR_false);
        }

        if (outputItems.Count > 0)
        {
            writer.WriteStartElement("output_settings");
            writer.WriteAttributeString("type", "array");

            foreach (RequestOutputItem item in outputItems)
            {
                item.ToJsonObject(writer);
            }

            writer.WriteEndElement(); // array
        }

        ScanUtils.writeElementArray(writer, "retrieve_caps", retrieveCaps);
        ScanUtils.writeElementArray(writer, "retrieve_extended_image_attrs", retrieveExtendedImageAttributes);

        if (imageFiles != null && imageFiles.Count > 0)
        {
            ScanUtils.writeElementArray(writer, "image_files", imageFiles);
        }

        if(useAspriseDialog) {
            writer.WriteElementString("use_asprise_dialog", useAspriseDialog ? ScanUtils.STR_true : ScanUtils.STR_false);
        }

        if(dialogTitle != null) {
            writer.WriteElementString("dialog_title", dialogTitle);
        }
        if(dialogWidth > 0) {
            ScanUtils.writeElementInt(writer, "dialog_width", dialogWidth);
        }
        if(dialogHeight > 0) {
            ScanUtils.writeElementInt(writer, "dialog_height", dialogWidth);
        }
        
        writer.WriteElementString("show_scanner_ui", showScannerUI ? ScanUtils.STR_true : ScanUtils.STR_false);
        writer.WriteElementString("modal_scanner_ui", modalScannerUI ? ScanUtils.STR_true : ScanUtils.STR_false);
        if(sourceName != null) {
            writer.WriteElementString("source_name", sourceName);
        }

        writer.WriteEndElement();
    }
    
    /// <summary>
    /// Request ID
    /// </summary>
    /// <returns></returns>
    public String GetId() {
        return id;
    }

    /// <summary>
    /// Request ID
    /// </summary>
    /// <param name="id"></param>
    /// <returns></returns>
    public Request SetId(String id) {
        this.id = id;
        return this;
    }

    /// <summary>
    /// By default images will be processed after each scan; Set to true to process images after all scans.
    /// Processing images after all scan could speed up the scanning process.
    /// </summary>
    /// <returns></returns>
    public Boolean IsProcessImagesAfterAllScans() {
        return processImagesAfterAllScans;
    }

    /// <summary>
    /// By default images will be processed after each scan; Set to true to process images after all scans.
    /// Processing images after all scan could speed up the scanning process.
    /// </summary>
    /// <param name="processImagesAfterAllScan"></param>
    /// <returns></returns>
    public Request SetProcessImagesAfterAllScans(Boolean processImagesAfterAllScan) {
        this.processImagesAfterAllScans = processImagesAfterAllScan;
        return this;
    }

    /// <summary>
    /// Whether to prompt scan more dialog.
    /// </summary>
    /// <returns></returns>
    public Boolean isPromptScanMore() {
        return promptScanMore;
    }

    /// <summary>
    /// Whether to prompt scan more dialog.
    /// </summary>
    /// <param name="promptScanMore"></param>
    /// <returns></returns>
    public Request SetPromptScanMore(Boolean promptScanMore) {
        this.promptScanMore = promptScanMore;
        return this;
    }

    /// <summary>
    /// Whether blank pages should be discard.
    /// </summary>
    /// <returns></returns>
    public Boolean IsDiscardBlankPages() {
        return discardBlankPages;
    }

    /// <summary>
    /// Whether blank pages should be discard.
    /// </summary>
    /// <param name="discardBlankPages"></param>
    /// <returns></returns>
    public Request SetDiscardBlankPages(Boolean discardBlankPages) {
        this.discardBlankPages = discardBlankPages;
        return this;
    }

    /// <summary>
    /// Ink coverage threshold to determine if a page is blank.
    /// </summary>
    /// <returns></returns>
    public decimal GetBlankPageThreshold() {
        return blankPageThreshold;
    }

    /// <summary>
    /// Ink coverage threshold to determine if a page is blank.
    /// </summary>
    /// <param name="blankPageThreshold"></param>
    /// <returns></returns>
    public Request SetBlankPageThreshold(decimal blankPageThreshold) {
        this.blankPageThreshold = blankPageThreshold;
        return this;
    }

    /// <summary>
    /// Percentage of margin to be excluded when calculating ink coverage.
    /// </summary>
    /// <returns></returns>
    public int GetBlankPageMarginPercent() {
        return blankPageMarginPercent;
    }

    /// <summary>
    /// Percentage of margin to be excluded when calculating ink coverage.
    /// </summary>
    /// <param name="blankPageMarginPercent"></param>
    /// <returns></returns>
    public Request SetBlankPageMarginPercent(int blankPageMarginPercent) {
        this.blankPageMarginPercent = blankPageMarginPercent;
        return this;
    }

    /// <summary>
    /// Whether barcodes should be recognized.
    /// </summary>
    /// <returns></returns>
    public Boolean IsRecognizeBarcodes() {
        return recognizeBarcodes;
    }

    /// <summary>
    /// Whether barcodes should be recognized.
    /// </summary>
    /// <param name="recognizeBarcodes"></param>
    /// <returns></returns>
    public Request SetRecognizeBarcodes(Boolean recognizeBarcodes) {
        this.recognizeBarcodes = recognizeBarcodes;
        return this;
    }

    public Boolean isOnWebServer() {
        return onWebServer;
    }

    public void SetOnWebServer(Boolean value) {
        onWebServer = value;
    }

    /// <summary>
    /// Map of cap name (could be appended with operation and priority, e.g., /RESET, /SET/3456).
    /// </summary>
    /// <returns></returns>
    public Dictionary<String, String> GetTwainCapSetting() {
        return twainCapSetting;
    }

    /// <summary>
    /// Set or reset twain capability.
    /// </summary>
    /// <param name="capCode">capCode e.g., TwainConstants#ICAP_PIXELTYPE</param>
    /// <param name="value">support raw value and twain constant expansion; support fall back values e.g., "TWPT_GRAY,TWPT_BW"</param>
    /// <param name="reset">true to set operation RESET instead of SET.</param>
    /// <returns></returns>
    public Request SetTwainCap(int capCode, Object value, Boolean reset) {
        this.twainCapSetting.Add(TwainConstants.getCapName(capCode) + (reset ? "/RESET" : ""), value == null ? null : value.ToString());
        return this;
    }

    /// <summary>
    /// Set twain capability.
    /// </summary>
    /// <param name="capCode">e.g., TwainConstants#ICAP_PIXELTYPE</param>
    /// <param name="value">support raw value and twain constant expansion; support fall back values e.g., "TWPT_GRAY,TWPT_BW"</param>
    /// <returns></returns>
    public Request SetTwainCap(int capCode, Object value) {
        return SetTwainCap(capCode, value, false);
    }

    /// <summary>
    /// Returns the cap value setting specified in the request as int or -1 if failed to convert to int.
    /// </summary>
    /// <param name="capCode"></param>
    /// <returns></returns>
    public int GetTwainCapValueAsInt(int capCode) {
        string capName = TwainConstants.getCapName(capCode);
        string value = twainCapSetting.ContainsKey(capName) ? twainCapSetting[capName] : null;
        if(value is String) {
            String sVal = (String)value;
            if(sVal.Contains(",")) {
                sVal = sVal.Substring(0, sVal.IndexOf(","));
            }
            if(sVal.Contains("TW")) {
                int intVal = TwainConstants.getConstantValue(sVal.Trim());
                return intVal;
            }
        }
        return ScanUtils.toInteger(value, -1);
    }

    /// <summary>
    /// Names of caps to be retrieved.
    /// </summary>
    /// <returns></returns>
    public List<String> GetRetrieveCaps() {
        return retrieveCaps;
    }

    /// <summary>
    /// Add a cap to be retrieved.
    /// </summary>
    /// <param name="capCode"></param>
    /// <returns></returns>
    public Request RetrieveCap(int capCode) {
        String capName = TwainConstants.getCapName(capCode);
        this.retrieveCaps.Add(capName == null ? capCode.ToString() : capName);
        return this;
    }

    /// <summary>
    /// Name of extended image attributes to be retrieved.
    /// </summary>
    /// <returns></returns>
    public List<String> GetRetrieveExtendedImageAttributes() {
        return retrieveExtendedImageAttributes;
    }

    /// <summary>
    /// Adds an extended image attribute to be retrieved.
    /// </summary>
    /// <param name="tweiCode"></param>
    /// <returns></returns>
    public Request AddRetrieveExtendedImageAttributes(int tweiCode) {
        String tweiName = TwainConstants.getTweiName(tweiCode);
        this.retrieveExtendedImageAttributes.Add(tweiName == null ? tweiCode.ToString() : tweiName);
        return this;
    }

    /// <summary>
    /// Output specifications.
    /// </summary>
    /// <returns></returns>
    public List<RequestOutputItem> GetOutputItems() {
        return outputItems;
    }

    /// <summary>
    /// Adds an output specification.
    /// </summary>
    /// <param name="outputItem"></param>
    /// <returns></returns>
    public Request AddOutputItem(RequestOutputItem outputItem) {
        this.outputItems.Add(outputItem);
        return this;
    }

    /// <summary>
    /// Adds output specifications.
    /// </summary>
    /// <param name="items"></param>
    /// <returns></returns>
    public Request AddOutputItems(List<RequestOutputItem> items) {
        this.outputItems.AddRange(items);
        return this;
    }

        /// <summary>
        /// Clears all output items.
        /// </summary>
        /// <returns></returns>
    public Request RemoveAllOutputItems() {
        this.outputItems.Clear();
        return this;
    }

        /// <summary>
        /// Returns all image files.
        /// </summary>
        /// <returns></returns>
    public List<string> GetImageFiles() {
        return imageFiles;
    }

    /// <summary>
    /// Adds an image file.
    /// </summary>
    /// <param name="imageFile"></param>
    /// <returns></returns>
    public Request AddImageFile(string imageFile) {
        if(imageFile == null || !File.Exists(imageFile)) {
            throw new ImagingException("File doesn't exist: " + imageFile);
        }
        if(!isAllASCII(imageFile)) {
            throw new ImagingException("File path contains non-ASCII char(s): " + imageFile);
        }

        this.imageFiles.Add(imageFile);
        return this;
    }

    /// <summary>
    /// Adds list of image files.
    /// </summary>
    /// <param name="imageFiles">list of image files or null.</param>
    /// <returns></returns>
    public Request addImageFiles(List<string> imageFiles) {
        if(imageFiles == null) {
            return this;
        }
        foreach(string file in imageFiles) {
            AddImageFile(file);
        }
        return this;
    }

    private static Boolean isAllASCII(String input) {
        Boolean isASCII = true;
        for (int i = 0; input != null && i < input.Length; i++) {
            int c = input[i];
            if (c > 0x7F) {
                isASCII = false;
                break;
            }
        }
        return isASCII;
    }

        /// <summary>
        /// Whether scanner UI should be shown.
        /// </summary>
        /// <param name="showScannerUI"></param>
    public void SetShowScannerUI(Boolean showScannerUI) {
        this.showScannerUI = showScannerUI;
    }

        /// <summary>
        /// Whether scanner UI should be modal.
        /// </summary>
        /// <param name="modalScannerUI"></param>
    public void SetModalScannerUI(Boolean modalScannerUI) {
        this.modalScannerUI = modalScannerUI;
    }

        /// <summary>
        /// The source name.
        /// </summary>
        /// <param name="sourceName"></param>
    public void SetSourceName(String sourceName) {
        this.sourceName = sourceName;
    }

        /// <summary>
        /// Whether scanner UI should be shown.
        /// </summary>
        /// <returns></returns>
    public Boolean GetShowScannerUI() {
        return showScannerUI;
    }

        /// <summary>
        /// Whether scanner UI should be modal.
        /// </summary>
        /// <returns></returns>
    public Boolean GetModalScannerUI() {
        return modalScannerUI;
    }

        /// <summary>
        /// Source name
        /// </summary>
        /// <returns></returns>
    public String GetSourceName() {
        return sourceName;
    }

    } // end class Request

    /// <summary>
    /// Specifies an output target in a scan request.
    /// </summary>
public class RequestOutputItem : ICloneable {
    /// <summary>
    /// Default thumbnail height
    /// </summary>
    public const int DEFAULT_THUMBNAIL_HEIGHT = 200;

    /// <summary>
    /// Default JPEG quality.
    /// </summary>
    public const int DEFAULT_JPEG_QUALITY = 80;

    /// <summary>
    /// Output type, any of AspriseImaging#OUTPUT_SAVE
    /// </summary>
    String type;

    /// <summary>
    /// Image format, any of AspriseImaging#FORMAT_JPG ...
    /// </summary>
    String format;

    /// <summary>
    /// Only for thumbnail output types; default value: 200.
    /// </summary>
    int thumbnailHeight = DEFAULT_THUMBNAIL_HEIGHT;

    /// <summary>
    /// Target output file path, may contain variable for expansion, e.g., C:\\tmp\\${TMS}${EXT}.
    /// </summary>
    String savePath;

    /// <summary>
    /// Only for JPG format; default value: 80
    /// </summary>
    int jpegQuality = DEFAULT_JPEG_QUALITY;

    /// <summary>
    /// Any of AspriseImaging#TIFF_COMPRESSION_CCITT_G4 ... Only for TIF format; default value: (null)
    /// </summary>
    String tiffCompression;

    /// <summary>
    /// Force to save images into individual single-image tiff files.
    /// </summary>
    Boolean tiffForceSinglePage;

    /// <summary>
    /// Force to output black/white PDF to save space.
    /// </summary>
    Boolean pdfForceBlackWhite;

    /// <summary>
    /// Set to true to output PDF/A compliant PDF.
    /// </summary>
    Boolean pdfaCompliant;

    /// <summary>
    /// PDF owner password
    /// </summary>
    String pdfOwnerPassword;

    /// <summary>
    /// PDF user password
    /// </summary>
    String pdfUserPassword;

    /// <summary>
    /// Optionally prints a line of text at the bottom of the first page.
    /// </summary>
    String pdfTextLine;

    /// <summary>
    /// Exif tags, key should be any of AspriseImaging#EXIF_NAME_DocumentName ...
    /// </summary>
    Dictionary<String, String> exifTags = new Dictionary<String, String>();

    /// <summary>
    /// Perform upload after all scans. Default is true.
    /// </summary>
    Boolean uploadAfterAllDone = true;

    /// <summary>
    /// Set to true to upload one image per HTTP request; otherwise upload all images in one HTTP request. Default: false.
    /// </summary>
    Boolean uploadOneByOne = false;

    /// <summary>
    /// must be set when type is AspriseImaging#OUTPUT_UPLOAD or AspriseImaging#OUTPUT_UPLOAD_THUMB.
    /// </summary>
    UploadSetting uploadSetting;

    public object Clone() {
        RequestOutputItem clone = (RequestOutputItem) this.MemberwiseClone();
        clone.exifTags = new Dictionary<string,string>(exifTags);
        clone.uploadSetting = uploadSetting == null ? null : (UploadSetting) uploadSetting.Clone();
        return clone;
    }

    public RequestOutputItem() {
    }

    /// <summary>
    /// Constructor.
    /// </summary>
    /// <param name="type">Any of AspriseImaging.OUTPUT_SAVE</param>
    /// <param name="format">Any of AspriseImaging.FORMAT_JPG</param>
    public RequestOutputItem(String type, String format) {
        SetType(type);
        SetFormat(format);
    }

    /// <summary>
    /// Deserialize from JSON.
    /// </summary>
    /// <param name="xe"></param>
    /// <returns></returns>
    public static RequestOutputItem FromJsonMap(XElement xe) {
        if(! xe.Attribute("type").Value.Equals("object")) {
            throw new ImagingException("Expecting type=object, actual: " + xe);
        }

        RequestOutputItem item = new RequestOutputItem();
        item.type = ScanUtils.xeElementValueAsString(xe, "type");
        item.format = ScanUtils.xeElementValueAsString(xe, "format");
        item.thumbnailHeight = ScanUtils.xeElementValueAsInt(xe, "thumbnail_height", DEFAULT_THUMBNAIL_HEIGHT);
        item.savePath = ScanUtils.xeElementValueAsString(xe, "save_path");
        item.jpegQuality = ScanUtils.xeElementValueAsInt(xe, "jpeg_quality", DEFAULT_JPEG_QUALITY);
        item.tiffCompression = ScanUtils.xeElementValueAsString(xe, "tiff_compression");
        item.tiffForceSinglePage =
            ScanUtils.equalsIgnoreCase(ScanUtils.xeElementValueAsString(xe, "tiff_force_single_page", "false"), "true");
        item.pdfForceBlackWhite = ScanUtils.xeElementValueAsString(xe, "pdf_force_black_white", "false").Equals("true");
        item.pdfaCompliant = ScanUtils.xeElementValueAsString(xe, "pdfa_compliant", "false").Equals("true");
        item.pdfOwnerPassword = ScanUtils.xeElementValueAsString(xe, "pdf_owner_password");
        item.pdfUserPassword = ScanUtils.xeElementValueAsString(xe, "pdf_user_password");
        item.pdfTextLine = ScanUtils.xeElementValueAsString(xe, "pdf_text_line");

        item.exifTags = ScanUtils.toDictStringString(ScanUtils.xeElementObjectToDictionary(xe.Element("exif")));

        item.uploadAfterAllDone =
            ScanUtils.equalsIgnoreCase(ScanUtils.xeElementValueAsString(xe, "upload_after_all_done", "true"), "true"); // default is true
        item.uploadOneByOne = 
            ScanUtils.equalsIgnoreCase(ScanUtils.xeElementValueAsString(xe,  "upload_one_by_one", "false"), "true");
        item.uploadSetting = UploadSetting.fromJsonMap(xe.Element("upload_target"));
        return item;
    }

    /// <summary>
    /// Serialize to JSON.
    /// </summary>
    /// <param name="writer"></param>
    public void ToJsonObject(XmlDictionaryWriter writer) {
        // validate first
        if (AspriseImaging.OUTPUT_SAVE.Equals(type) || AspriseImaging.OUTPUT_SAVE_THUMB.Equals(type))
        {
            if (savePath == null || savePath.Trim().Length == 0)
            {
                throw new ArgumentException("savePath must be set for " + type);
            }
        }
        else if (AspriseImaging.OUTPUT_UPLOAD.Equals(type) || AspriseImaging.OUTPUT_UPLOAD_THUMB.Equals(type))
        {
            if (uploadSetting == null)
            {
                throw new ArgumentException("uploadSetting must be set for " + type);
            }
        }

        writer.WriteStartElement("item");
        writer.WriteAttributeString("type", "object");

        writer.WriteElementString("type", type);
        writer.WriteElementString("format", format);
        ScanUtils.writeElementInt(writer, "thumbnail_height", thumbnailHeight);
        if(savePath != null) {
            writer.WriteElementString("save_path", savePath);
        }

        if (jpegQuality > 0)
        {
            ScanUtils.writeElementInt(writer, "jpeg_quality", jpegQuality);
        }

        if (tiffCompression != null)
        {
            writer.WriteElementString("tiff_compression", tiffCompression);
        }

        if (tiffForceSinglePage)
        {
            writer.WriteElementString("tiff_force_single_page", tiffForceSinglePage ? "true" : "false");
        }

        if (pdfForceBlackWhite) {
            writer.WriteElementString("pdf_force_black_white", pdfForceBlackWhite ? "true" : "false");
        }

        if (pdfaCompliant)
        {
            writer.WriteElementString("pdfa_compliant", pdfaCompliant ? "true" : "false");
        }

        if(pdfOwnerPassword != null) {
            writer.WriteElementString("pdf_owner_password", pdfOwnerPassword);
        }

        if (pdfUserPassword != null)
        {
            writer.WriteElementString("pdf_user_password", pdfUserPassword);
        }

        if (pdfTextLine != null)
        {
            writer.WriteElementString("pdf_text_line", pdfTextLine);
        }

        ScanUtils.writeElementDict(writer, "exif", exifTags);

        writer.WriteElementString("pdfa_compliant", pdfaCompliant ? "true" : "false");

        writer.WriteElementString("upload_after_all_done", uploadAfterAllDone ? "true" : "false");
        writer.WriteElementString("upload_one_by_one", uploadOneByOne ? "true" : "false");

        if (uploadSetting != null)
        {
            uploadSetting.toJsonObject(writer);
        }

        writer.WriteEndElement(); // root
    }

    /// <summary>
    /// Output type, any of AspriseImaging#OUTPUT_SAVE ...
    /// </summary>
    /// <returns></returns>
    public String GetOutputType() {
        return type;
    }

    /// <summary>
    /// Output type, any of AspriseImaging#OUTPUT_SAVE ...
    /// </summary>
    /// <param name="type"></param>
    /// <returns></returns>
    public RequestOutputItem SetType(String type) {
        if (AspriseImaging.OUTPUT_SAVE.Equals(type) || AspriseImaging.OUTPUT_SAVE_THUMB.Equals(type)
                || AspriseImaging.OUTPUT_RETURN_BASE64.Equals(type) || AspriseImaging.OUTPUT_RETURN_BASE64_THUMB.Equals(type)
                || AspriseImaging.OUTPUT_UPLOAD.Equals(type) || AspriseImaging.OUTPUT_UPLOAD_THUMB.Equals(type))
        {
            // ok
        } else {
            throw new ImagingException("Output type: " + type);
        }
        this.type = type;
        return this;
    }

    /// <summary>
    /// Image format, any of AspriseImaging#FORMAT_JPG
    /// </summary>
    /// <returns></returns>
    public String GetFormat() {
        return format;
    }

    /// <summary>
    /// Image format, any of AspriseImaging#FORMAT_JPG*/
    /// </summary>
    /// <param name="format"></param>
    /// <returns></returns>
    public RequestOutputItem SetFormat(String format) {
        if (AspriseImaging.FORMAT_JPG.Equals(format) || AspriseImaging.FORMAT_PDF.Equals(format) || AspriseImaging.FORMAT_TIF.Equals(format)
                || AspriseImaging.FORMAT_PNG.Equals(format) || AspriseImaging.FORMAT_BMP.Equals(format))
        {
            // ok
        } else {
            throw new ImagingException("Image format: " + format);
        }
        this.format = format;
        return this;
    }

    /// <summary>
    /// Only for thumbnail output types; default value: 200.
    /// </summary>
    /// <returns></returns>
    public int GetThumbnailHeight() {
        return thumbnailHeight;
    }

    /// <summary>
    /// Only for thumbnail output types; default value: 200.
    /// </summary>
    /// <param name="thumbnailHeight"></param>
    /// <returns></returns>
    public RequestOutputItem SetThumbnailHeight(int thumbnailHeight) {
        this.thumbnailHeight = thumbnailHeight;
        return this;
    }

    /// <summary>
    /// Target output file path, may contain variable for expansion, e.g., C:\\tmp\\${TMS}${EXT}. 
    /// </summary>
    /// <returns></returns>
    public String GetSavePath() {
        return savePath;
    }

    /// <summary>
    /// Target output file path, may contain variable for expansion, e.g., C:\\tmp\\${TMS}${EXT}.
    /// </summary>
    /// <param name="savePath"></param>
    /// <returns></returns>
    public RequestOutputItem SetSavePath(String savePath) {
        this.savePath = savePath;
        return this;
    }

    /// <summary>
    /// Only for JPG format; default value: 80
    /// </summary>
    /// <returns></returns>
    public int GetJpegQuality() {
        return jpegQuality;
    }

    /// <summary>
    /// Only for JPG format; default value: 80
    /// </summary>
    /// <param name="jpegQuality"></param>
    /// <returns></returns>
    public RequestOutputItem SetJpegQuality(int jpegQuality) {
        this.jpegQuality = jpegQuality;
        return this;
    }

    /// <summary>
    /// Any of AspriseImaging#TIFF_COMPRESSION_CCITT_G4 ... Only for TIF format; default value: (null)
    /// </summary>
    /// <returns></returns>
    public String GetTiffCompression() {
        return tiffCompression;
    }

    /// <summary>
    /// Any of AspriseImaging#TIFF_COMPRESSION_CCITT_G4 ... Only for TIF format; default value: (null)
    /// </summary>
    /// <param name="tiffCompression"></param>
    /// <returns></returns>
    public RequestOutputItem SetTiffCompression(String tiffCompression) {
        this.tiffCompression = tiffCompression;
        return this;
    }

    /// <summary>
    /// Force to save images into individual single-image tiff files.
    /// </summary>
    /// <returns></returns>
    public Boolean IsTiffForceSinglePage() {
        return tiffForceSinglePage;
    }

    /// <summary>
    /// Force to save images into individual single-image tiff files.
    /// </summary>
    /// <param name="tiffForceSinglePage"></param>
    /// <returns></returns>
    public RequestOutputItem setTiffForceSinglePage(Boolean tiffForceSinglePage) {
        this.tiffForceSinglePage = tiffForceSinglePage;
        return this;
    }

    /// <summary>
    /// Force to output black/white PDF to save space.
    /// </summary>
    /// <returns></returns>
    public Boolean IsPdfForceBlackWhite() {
        return pdfForceBlackWhite;
    }

    /// <summary>
    /// Force to output black/white PDF to save space.
    /// </summary>
    /// <param name="pdfForceBlackWhite"></param>
    /// <returns></returns>
    public RequestOutputItem SetPdfForceBlackWhite(Boolean pdfForceBlackWhite) {
        this.pdfForceBlackWhite = pdfForceBlackWhite;
        return this;
    }

    /// <summary>
    /// Set to true to output PDF/A compliant PDF.
    /// </summary>
    /// <returns></returns>
    public Boolean IsPdfaCompliant() {
        return pdfaCompliant;
    }

    /// <summary>
    /// Set to true to output PDF/A compliant PDF.
    /// </summary>
    /// <param name="pdfaCompliant"></param>
    /// <returns></returns>
    public RequestOutputItem SetPdfaCompliant(Boolean pdfaCompliant) {
        this.pdfaCompliant = pdfaCompliant;
        return this;
    }

    /// <summary>
    /// PDF owner password
    /// </summary>
    /// <returns></returns>
    public String GetPdfOwnerPassword() {
        return pdfOwnerPassword;
    }

    /// <summary>
    /// PDF owner password
    /// </summary>
    /// <param name="pdfOwnerPassword"></param>
    /// <returns></returns>
    public RequestOutputItem SetPdfOwnerPassword(String pdfOwnerPassword) {
        this.pdfOwnerPassword = pdfOwnerPassword;
        return this;
    }

    /// <summary>
    /// PDF user password
    /// </summary>
    /// <returns></returns>
    public String GetPdfUserPassword() {
        return pdfUserPassword;
    }

    /// <summary>
    /// PDF user password
    /// </summary>
    /// <param name="pdfUserPassword"></param>
    /// <returns></returns>
    public RequestOutputItem SetPdfUserPassword(String pdfUserPassword) {
        this.pdfUserPassword = pdfUserPassword;
        return this;
    }

    /// <summary>
    /// Optionally prints a line of text at the bottom of the first page.
    /// </summary>
    /// <returns></returns>
    public String GetPdfTextLine() {
        return pdfTextLine;
    }

    /// <summary>
    /// Optionally prints a line of text at the bottom of the first page.
    /// </summary>
    /// <param name="pdfTextLine"></param>
    /// <returns></returns>
    public RequestOutputItem SetPdfTextLine(String pdfTextLine) {
        this.pdfTextLine = pdfTextLine;
        return this;
    }

    /// <summary>
    /// Exif tags, key should be any of AspriseImaging#EXIF_NAME_DocumentName
    /// </summary>
    /// <returns></returns>
    public Dictionary<String, String> GetExifTags() {
        return exifTags;
    }

    /// <summary>
    /// Exif tags, key should be any of AspriseImaging#EXIF_NAME_DocumentName
    /// </summary>
    /// <param name="tagName"></param>
    /// <param name="value"></param>
    /// <returns></returns>
    public RequestOutputItem AddExifTag(String tagName, String value) {
        this.exifTags.Add(tagName, value);
        return this;
    }

    /// <summary>
    /// Perform upload after all scans. Default is true.
    /// </summary>
    /// <returns></returns>
    public Boolean IsUploadAfterAllDone() {
        return uploadAfterAllDone;
    }

    /// <summary>
    /// Perform upload after all scans. Default is true.
    /// </summary>
    /// <param name="uploadAfterAllDone"></param>
    /// <returns></returns>
    public RequestOutputItem SetUploadAfterAllDone(Boolean uploadAfterAllDone) {
        this.uploadAfterAllDone = uploadAfterAllDone;
        return this;
    }

    /// <summary>
    /// Set to true to upload one image per HTTP request; otherwise upload all images in one HTTP request. Default: false.
    /// </summary>
    /// <returns></returns>
    public Boolean IsUploadOneByOne() {
        return uploadOneByOne;
    }

    /// <summary>
    /// Set to true to upload one image per HTTP request; otherwise upload all images in one HTTP request. Default: false.
    /// </summary>
    /// <param name="uploadOneByOne"></param>
    /// <returns></returns>
    public RequestOutputItem SetUploadOneByOne(Boolean uploadOneByOne) {
        this.uploadOneByOne = uploadOneByOne;
        return this;
    }

    /// <summary>
    /// must be set when type is AspriseImaging#OUTPUT_UPLOAD or AspriseImaging#OUTPUT_UPLOAD_THUMB.
    /// </summary>
    /// <returns></returns>
    public UploadSetting GetUploadSetting() {
        return uploadSetting;
    }

    /// <summary>
    /// must be set when type is AspriseImaging#OUTPUT_UPLOAD AspriseImaging#OUTPUT_UPLOAD_THUMB.
    /// </summary>
    /// <param name="uploadSetting"></param>
    /// <returns></returns>
    public RequestOutputItem SetUploadSetting(UploadSetting uploadSetting) {
        this.uploadSetting = uploadSetting;
        return this;
    }
}

    /// <summary>
    /// Upload settings - url is the only mandatory parameter.
    /// </summary>
    public class UploadSetting : ICloneable
    {
        /// <summary>
        /// Default max number of retries.
        /// </summary>
    public const int DEFAULT_MAX_RETRIES = 2;
        /// <summary>
        /// Default max operation time allowed in seconds.
        /// </summary>
    public const int DEFAULT_MAX_OPERATION_TIME = 1200;
        /// <summary>
        /// Default POST name for uploaded files 
        /// </summary>
    public const String DEFAULT_POST_FILE_FIELD_NAME = "asprise_scans";

    /** Target URL */
    String url;

    /** Max number of retries. Default is 2. */
    int maxRetries = DEFAULT_MAX_RETRIES;

    /** Max HTTP operation time in seconds; default: 1200 (20 mins). */
    int maxOperationTime = DEFAULT_MAX_OPERATION_TIME;

    /** Field name for files. */
    String postFileFieldName = DEFAULT_POST_FILE_FIELD_NAME;

    /** Additional files to be posted along */
    List<String> postFiles = new List<String>();

    /** Any additional post fields */
    Dictionary<String, String> postFields = new Dictionary<String, String>();

    /** e.g., "name=Asprise; domain=asprise.com" */
    String cookies;

    /** Authentication, e.g., "user:pass" */
    String auth;

    /** e.g., "Referer: http://asprise.com" */
    List<String> headers = new List<String>();

    /** Saves response to the given file instead of returning response. */
    String saveResponseToFile;

    /** Log file to record HTTP operations. */
    String logToFile;

        /// <summary>
        /// Deep clones.
        /// </summary>
        /// <returns></returns>
    public object Clone() {
        UploadSetting clone = (UploadSetting) this.MemberwiseClone();
        clone.postFiles = new List<String>(postFiles);
        clone.postFields = new Dictionary<String, String>(postFields);
        clone.headers = new List<String>(headers);
        return clone;
    }

    public UploadSetting(String url) {
        this.url = url;
    }

    public UploadSetting() {
    }

    /// <summary>
    /// Target URL
    /// </summary>
    /// <returns></returns>
    public String GetUrl() {
        return url;
    }

    /// <summary>
    /// Target URL
    /// </summary>
    /// <param name="url"></param>
    /// <returns></returns>
    public UploadSetting SetUrl(String url) {
        this.url = url;
        return this;
    }

    /// <summary>
    /// Max number of retries. Default is 2.
    /// </summary>
    /// <returns></returns>
    public int GetMaxRetries() {
        return maxRetries;
    }

    /// <summary>
    /// Max number of retries. Default is 2.
    /// </summary>
    /// <param name="maxRetries"></param>
    /// <returns></returns>
    public UploadSetting SetMaxRetries(int maxRetries) {
        this.maxRetries = maxRetries;
        return this;
    }

    /// <summary>
    /// Max HTTP operation time in seconds; default: 1200 (20 mins).
    /// </summary>
    /// <returns></returns>
    public int GetMaxOperationTime() {
        return maxOperationTime;
    }

    /// <summary>
    /// Max HTTP operation time in seconds; default: 1200 (20 mins).
    /// </summary>
    /// <param name="maxOperationTime"></param>
    /// <returns></returns>
    public UploadSetting SetMaxOperationTime(int maxOperationTime) {
        this.maxOperationTime = maxOperationTime;
        return this;
    }

    /// <summary>
    /// Field name for files.
    /// </summary>
    /// <returns></returns>
    public String GetPostFileFieldName() {
        return postFileFieldName;
    }

    /// <summary>
    /// Field name for files.
    /// </summary>
    /// <param name="postFileFieldName"></param>
    /// <returns></returns>
    public UploadSetting SetPostFileFieldName(String postFileFieldName) {
        this.postFileFieldName = postFileFieldName;
        return this;
    }

    /// <summary>
    /// Additional files to be posted along
    /// </summary>
    /// <returns></returns>
    public List<String> GetPostFiles() {
        return postFiles;
    }

    /// <summary>
    /// Additional files to be posted along.
    /// </summary>
    /// <param name="postFile"></param>
    /// <returns></returns>
    public UploadSetting AddPostFiles(String postFile) {
        this.postFiles.Add(postFile);
        return this;
    }

    /// <summary>
    /// Any additional post fields.
    /// </summary>
    /// <returns></returns>
    public Dictionary<String, String> GetPostFields() {
        return postFields;
    }

    /// <summary>
    /// Any additional post fields.
    /// </summary>
    /// <param name="name"></param>
    /// <param name="value"></param>
    /// <returns></returns>
    public UploadSetting AddPostField(String name, String value) {
        this.postFields.Add(name, value);
        return this;
    }

    /// <summary>
    /// e.g., "name=Asprise; domain=asprise.com"
    /// </summary>
    /// <returns></returns>
    public String GetCookies() {
        return cookies;
    }

    /// <summary>
    /// e.g., "name=Asprise; domain=asprise.com"
    /// </summary>
    /// <param name="cookies"></param>
    /// <returns></returns>
    public UploadSetting SetCookies(String cookies) {
        this.cookies = cookies;
        return this;
    }

    /// <summary>
    /// Authentication, e.g., "user:pass"
    /// </summary>
    /// <returns></returns>
    public String GetAuth() {
        return auth;
    }

    /// <summary>
    /// Authentication, e.g., "user:pass".
    /// </summary>
    /// <param name="auth"></param>
    /// <returns></returns>
    public UploadSetting SetAuth(String auth) {
        this.auth = auth;
        return this;
    }

    /// <summary>
    /// e.g., "Referer: http://asprise.com"
    /// </summary>
    /// <returns></returns>
    public List<String> GetHeaders() {
        return headers;
    }

    /// <summary>
    /// e.g., "Referer: http://asprise.com"
    /// </summary>
    /// <param name="header"></param>
    /// <returns></returns>
    public UploadSetting AddHeaders(String header) {
        this.headers.Add(header);
        return this;
    }

    /// <summary>
    /// Saves response to the given file instead of returning response.
    /// </summary>
    /// <returns></returns>
    public string GetSaveResponseToFile() {
        return saveResponseToFile;
    }

    /// <summary>
    /// Saves response to the given file instead of returning response.
    /// </summary>
    /// <param name="saveResponseToFile"></param>
    /// <returns></returns>
    public UploadSetting SetSaveResponseToFile(String saveResponseToFile) {
        this.saveResponseToFile = saveResponseToFile;
        return this;
    }

    /// <summary>
    /// Log file to record HTTP operations.
    /// </summary>
    /// <returns></returns>
    public String GetLogToFile() {
        return logToFile;
    }

    /// <summary>
    /// Log file to record HTTP operations.
    /// </summary>
    /// <param name="httpLogFile"></param>
    /// <returns></returns>
    public UploadSetting SetLogToFile(String httpLogFile) {
        this.logToFile = httpLogFile;
        return this;
    }

        /// <summary>
        /// Deserialize from JSON.
        /// </summary>
        /// <param name="xe"></param>
        /// <returns></returns>
    public static UploadSetting fromJsonMap(XElement xe) {
        if (xe == null)
        {
            return null;
        }

        if(! xe.Name.LocalName.Equals("upload_target")) {
            throw new ImagingException("Expecting upload_target, actual: " + xe);
        }
        if(! xe.Attribute("type").Value.Equals("object")) {
            throw new ImagingException("Expecting type=object, actual: " + xe);
        }

        UploadSetting setting = new UploadSetting();
        setting.url = ScanUtils.xeElementValueAsString(xe, "url");
        setting.maxRetries = ScanUtils.xeElementValueAsInt(xe, "max_retries", DEFAULT_MAX_RETRIES);
        Dictionary<string, object> postFieldsInJson = ScanUtils.xeElementObjectToDictionary(xe.Element("post_fields"));
        if(postFieldsInJson != null) {
            foreach (KeyValuePair<string, object> pair in postFieldsInJson)
            {
                setting.postFields.Add(pair.Key, pair.Value == null ? "" : pair.Value.ToString());
            }
        }

        setting.postFileFieldName = ScanUtils.xeElementValueAsString(xe, "post_file_field_name", DEFAULT_POST_FILE_FIELD_NAME);

        setting.postFiles = ScanUtils.toListOfString(ScanUtils.xeElementArrayToList(xe.Element("post_files")));

        setting.cookies = ScanUtils.xeElementValueAsString(xe, "cookies");
        setting.auth = ScanUtils.xeElementValueAsString(xe, "auth");

        setting.headers = ScanUtils.toListOfString(ScanUtils.xeElementArrayToList(xe.Element("headers")));

        setting.logToFile = ScanUtils.xeElementValueAsString(xe, "log_file");
        setting.saveResponseToFile = ScanUtils.xeElementValueAsString(xe, "to_file");
        if(ScanUtils.equalsIgnoreCase(setting.saveResponseToFile, "null"))
        {
            setting.saveResponseToFile = null;
        }

        setting.maxOperationTime = ScanUtils.xeElementValueAsInt(xe, "max_operation_time", DEFAULT_MAX_OPERATION_TIME);
        return setting;
    }

        /// <summary>
        /// Serializes to JSON.
        /// </summary>
        /// <param name="writer"></param>
        /// <param name="elementName"></param>
    public void toJsonObject(XmlDictionaryWriter writer, string elementName = "upload_target")
    {
        writer.WriteStartElement(elementName);
        writer.WriteAttributeString("type", "object");
      
        writer.WriteElementString("url", url);
        ScanUtils.writeElementInt(writer, "max_retries", maxRetries);
        ScanUtils.writeElementDict(writer, "post_fields", postFields);
        writer.WriteElementString("post_file_field_name", postFileFieldName);
        ScanUtils.writeElementArray(writer, "post_files", postFiles);

        if (cookies != null && cookies.Trim().Length > 0)
        {
            writer.WriteElementString("cookies", cookies);
        }

        if (auth != null && auth.Trim().Length > 0)
        {
            writer.WriteElementString("cookies", cookies);
        }

        ScanUtils.writeElementArray(writer, "headers", headers);

        if (logToFile != null && logToFile.Trim().Length > 0)
        {
            writer.WriteElementString("log_file", logToFile);
        }

        if (saveResponseToFile != null && saveResponseToFile.Trim().Length > 0)
        {
            writer.WriteElementString("to_file", saveResponseToFile);
        }

        ScanUtils.writeElementInt(writer, "max_operation_time", maxOperationTime);

        writer.WriteEndElement(); // end root
    }
} // end: class Request
} // end: namespace asprise_imaging_api
