﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml;
using System.Xml.Linq;
using System.Xml.XPath;
using System.IO;
using System.Drawing;

namespace asprise_imaging_api
{
    /// <summary>
    /// Represents a scan result.
    /// </summary>
    public class Result
    {

    /** Source device */
    protected String device;
    /** Number of images acquired */
    protected int imageCount = -1;
    /** Last transfer code, e.g., {@linkplain TwainConstants#TWRC_SUCCESS}. */
    protected int lastTransferResultCode = -1;

    protected String requestId;
    protected String responseId;

    // String rawResult; // could boost memory footprint, so we disable it here.

    protected List<ResultImageItem> images = new List<ResultImageItem>();

    protected List<ResultOutputItem> outputItems = new List<ResultOutputItem>();

    public Result(XElement xe) {
        device = ScanUtils.xeElementValueAsString(xe, "device");
        imageCount = ScanUtils.xeElementValueAsInt(xe, "image_count", -1);
        lastTransferResultCode = TwainConstants.getTwrcCode(ScanUtils.xeElementValueAsString(xe, "last_transfer_rc"));
        requestId = ScanUtils.xeElementValueAsString(xe, "request_id");
        responseId = ScanUtils.xeElementValueAsString(xe, "response_id");

        XElement xeImages = xe.Element("images");
        if(xeImages != null) {
            foreach(XElement xeImg in xeImages.Elements()) {
                images.Add(new ResultImageItem(xeImg));
            }
        }

        XElement xeOutputs = xe.Element("output");
        if(xeOutputs != null) {
            foreach(XElement xeOut in xeOutputs.Elements()) {
                outputItems.Add(new ResultOutputItem(xeOut));
            }
        }
    }

    /// <summary>
    /// Creates Result if the resultJson is not null.
    /// </summary>
    /// <param name="resultJson"></param>
    /// <returns></returns>
    public static Result fromJson(String resultJson) {
        if(resultJson == null) {
            return null;
        }
        return new Result(ScanUtils.parseJson(resultJson));
    }

        /// <summary>
        /// Deserializes from JSON.
        /// </summary>
        /// <param name="xe"></param>
        /// <returns>Result deserialized or null if the argement passed is null.</returns>
    public static Result createScanResult(XElement xe) {
        if(xe == null) {
            return null;
        } else {
            return new Result(xe);
        }
    }

    /// <summary>
    /// The source device from which images are acquired.
    /// </summary>
    public string Device
    {
        get
        {
            return device;
        }
    }

    /// <summary>
    /// Total number of images acquired.
    /// </summary>
    public int ImageCount
    {
        get
        {
            return imageCount;
        }
    }

    /// <summary>
    /// Last transfer result code, e.g., TwainConstants#TWRC_SUCCESS.
    /// </summary>
    public int LastTransferResultCode
    {
        get
        {
            return lastTransferResultCode;
        }
    }

    /// <summary>
    /// Whether user pressed 'Cancel'.
    /// </summary>
    /// <returns></returns>
    public Boolean isUserCancelled() {
        return lastTransferResultCode == TwainConstants.TWRC_CANCEL;
    }


    /// <summary>
    /// Lists of images acquired.
    /// </summary>
    /// <returns></returns>
    public List<ResultImageItem> GetImageItems() {
        return images;
    }

    /// <summary>
    /// List of all output items.
    /// </summary>
    /// <returns></returns>
    public List<ResultOutputItem> GetOutputItems() {
        return outputItems;
    }

    /// <summary>
    /// Finds the first output item of the given type.
    /// </summary>
    /// <param name="outputType">any of Imaging.OUTPUT_SAVE ...</param>
    /// <returns>the first output item of the list sorted by record count descending then format (JPG, PNG, BMP, TIF, PDF).</returns>
    public ResultOutputItem GetOutputItem(String outputType) {
        foreach (ResultOutputItem item in outputItems) { 
            if(item == null) {
                continue;
            }
            if(outputType.Equals(item.GetOutputType())) {
                return item;
            }
        }
        return null;
    }

        private static int compareRoiByRecords(ResultOutputItem o1, ResultOutputItem o2) {
            int recordDiff = o2.GetOutputRecordCount() - o1.GetOutputRecordCount();
            if(recordDiff != 0) {
                return recordDiff;
            }
            return GetFileFormatRanking(o1.GetFormat()) - GetFileFormatRanking(o2.GetFormat());
        }

    /// <summary>
    /// Finds the all output items of the given type.
    /// </summary>
    /// <param name="outputType">any of Imaging.OUTPUT_SAVE ...</param>
    /// <returns>All output items of the given type sorted by record count descending then format (JPG, PNG, BMP, TIF, PDF).</returns>
    public List<ResultOutputItem> GetOutputItems(String outputType) {
        List<ResultOutputItem> filtered = new List<ResultOutputItem>();
        foreach(ResultOutputItem outputItem in outputItems) {
            if(outputItem == null) {
                continue;
            }
            if(outputType.Equals(outputItem.GetOutputType())) {
                filtered.Add(outputItem);
            }
        }

        filtered.Sort(compareRoiByRecords);
        return filtered;
    }

    private static int GetFileFormatRanking(String format) {
        if(ScanUtils.equalsIgnoreCase(AspriseImaging.FORMAT_JPG, format)) {
            return 1;
        } else if(ScanUtils.equalsIgnoreCase(AspriseImaging.FORMAT_PNG, format)) {
            return 2;
        } else if(ScanUtils.equalsIgnoreCase(AspriseImaging.FORMAT_BMP, format)) {
            return 3;
        } else if(ScanUtils.equalsIgnoreCase(AspriseImaging.FORMAT_TIF, format)) {
            return 4;
        } else if(ScanUtils.equalsIgnoreCase(AspriseImaging.FORMAT_PDF, format)) {
            return 5;
        } else {
            return 6;
        }
    }

    /// <summary>
    /// Gets the list of image files acquired through OUTPUT_SAVE
    /// </summary>
    /// <returns>list of image files or null if no OUTPUT_SAVE record found.</returns>
    public List<String> GetImageFiles() {
        ResultOutputItem outputSave = GetOutputItem(AspriseImaging.OUTPUT_SAVE);
        return outputSave == null ? null : outputSave.GetFiles();
    }

    /// <summary>
    /// Gets the list of image files acquired through OUTPUT_SAVE_THUMB
    /// </summary>
    /// <returns>the list of image files or null if no OUTPUT_SAVE_THUMB record found.</returns>
    public List<String> GetThumbnailFiles() {
        ResultOutputItem outputSave = GetOutputItem(AspriseImaging.OUTPUT_SAVE_THUMB);
        return outputSave == null ? null : outputSave.GetFiles();
    }

    /// <summary>
    /// Gets the PDF file acquired through OUTPUT_SAVE
    /// </summary>
    /// <returns>the PDF file acquired through OUTPUT_SAVE or null if none is found.</returns>
    public String GetPdfFile() {
        List<ResultOutputItem> outputItems = GetOutputItems(AspriseImaging.OUTPUT_SAVE);
        foreach(ResultOutputItem item in outputItems) {
            if(ScanUtils.equalsIgnoreCase(AspriseImaging.FORMAT_PDF, item.GetFormat())) {
                List<String> files = item.GetFiles();
                return files != null && files.Count > 0 ? files[0] : null;
            }
        }
        return null;
    }

    /// <summary>
    /// Gets the TIFF file acquired through OUTPUT_SAVE
    /// </summary>
    /// <returns>the TIFF file acquired through OUTPUT_SAVE or null if none is found.</returns>
    public String GetTiffFile() {
        List<ResultOutputItem> outputItems = GetOutputItems(AspriseImaging.OUTPUT_SAVE);
        foreach(ResultOutputItem item in outputItems) {
            if(item.GetFormat() != null && item.GetFormat().ToLower().Contains(AspriseImaging.FORMAT_TIF)) {
                List<String> files = item.GetFiles();
                return files != null && files.Count > 0 ? files[0] : null;
            }
        }
        return null;
    }

    /// <summary>
    /// Gets the original image of the given index by searching the output in order: OUTPUT_RETURN_BASE64, OUTPUT_SAVE.
    /// This is a shortcut method, use GetOutputItems if you need more control.
    /// </summary>
    /// <param name="index">the image of the given index or null if not available.</param>
    /// <param name="includeSave">whether images acquired in SAVE output should be considered</param>
    /// <returns></returns>
    public Bitmap GetImage(int index, bool includeSave = true)
    {
        List<ResultOutputItem> outputReturnBase64s = GetOutputItems(AspriseImaging.OUTPUT_RETURN_BASE64);
        if(outputReturnBase64s != null && outputReturnBase64s.Count > 0) {
            return outputReturnBase64s[0].LoadImage(index);
        }
        if (includeSave)
        {
            List<ResultOutputItem> outputSaves = GetOutputItems(AspriseImaging.OUTPUT_SAVE);
            if (outputSaves != null && outputSaves.Count > 0)
            {
                return outputSaves[0].LoadImage(index);
            }
        }
        return null;
    }

    /// <summary>
    /// Returns all scanned images.
    /// </summary>
    /// <returns></returns>
    public List<Image> GetImages(bool includeSave = true) {
        List<Image> images = new List<Image>();
        for(int i = 0; i < imageCount; i++) {
            Bitmap bitmap = GetImage(i, includeSave);
            if (bitmap != null)
            {
                images.Add(bitmap);
            }
        }
        return images;
    }

    /// <summary>
    /// Gets the thumbnail image of the given index by searching the output in order: OUTPUT_RETURN_BASE64_THUMB, OUTPUT_SAVE_THUMB.
    /// This is a shortcut method, use {@linkplain #getOutputItems()} if you need more control.
    /// </summary>
    /// <param name="index"></param>
    /// <returns>the image of the given index or null if not available.</returns>
    public Bitmap GetThumbnail(int index) {
        List<ResultOutputItem> outputReturnBase64Thumbs = GetOutputItems(AspriseImaging.OUTPUT_RETURN_BASE64_THUMB);
        if(outputReturnBase64Thumbs != null && outputReturnBase64Thumbs.Count > 0) {
            return outputReturnBase64Thumbs[0].LoadImage(index);
        }
        List<ResultOutputItem> outputSaveThumbs = GetOutputItems(AspriseImaging.OUTPUT_SAVE_THUMB);
        if(outputSaveThumbs != null && outputSaveThumbs.Count > 0) {
            return outputSaveThumbs[0].LoadImage(index);
        }
        return null;
    }

    /// <summary>
    /// Returns the upload response if available.
    /// This is a shortcut method, use GetOutputItems() if you need more control.
    /// </summary>
    /// <returns>the upload response or null if not available.</returns>
    public String GetUploadResponse() {
        ResultOutputItem outputItemUpload = GetOutputItem(AspriseImaging.OUTPUT_UPLOAD);
        return outputItemUpload == null || outputItemUpload.GetOutputRecordCount() == 0 ?
                null : outputItemUpload.GetOutputRecords()[0];
    }

    /// <summary>
    /// Returns the thumbnail upload response if available.
    /// This is a shortcut method, use GetOutputItems() if you need more control.
    /// </summary>
    /// <returns>the thumbnail upload response or null if not available.</returns>
    public String GetThumbnailUploadResponse() {
        ResultOutputItem outputItemUpload = GetOutputItem(AspriseImaging.OUTPUT_UPLOAD_THUMB);
        return outputItemUpload == null || outputItemUpload.GetOutputRecordCount() == 0 ?
                null : outputItemUpload.GetOutputRecords()[0];
    }

    /// <summary>
    /// List of barcodes recognized
    /// </summary>
    /// <returns></returns>
    public List<String> GetBarcodes() {
        List<String> list = new List<String>();
        List<ResultImageItem> items = GetImageItems();
        for(int i = 0; i < items.Count; i++) {
            ResultImageItem item = items[i];
            List<Dictionary<string, object>> barcodes = item.GetBarcodes();
            if(barcodes == null || barcodes.Count == 0) {
                continue;
            }

            foreach(Dictionary<string, object> barcode in barcodes) {
                string s = mapToString(barcode);
                s = s + ", pageIndex=" + i;
                list.Add(s);
            }
        }
        return list;
    }

    static String mapToString(Dictionary<String, object> map) {
        StringBuilder sb = new StringBuilder();
        foreach(KeyValuePair<string, object> entry in map)
        {
            if(sb.Length > 0) {
                sb.Append(", ");
            }
            sb.Append(entry.Key).Append("=").Append(entry.Value);
        }

        return sb.ToString();
    }

    public override String ToString() {
        return "ScanResult{" +
                "device='" + device + '\'' +
                ", imageCount=" + imageCount +
                ", lastTransferResultCode=" + lastTransferResultCode +
                ", requestId='" + requestId + '\'' +
                ", responseId='" + responseId + '\'' +
                ", images=" + images +
                ", outputItems=" + outputItems +
                '}';
    }
    }

    /// <summary>
    /// Represents an output item.
    /// </summary>
public class ResultOutputItem {

    /// <summary>
    /// Any of AspriseImaging#OUTPUT_SAVE, etc.
    /// </summary>
    String type;

    /// <summary>
    /// Output image format
    /// </summary>
    String format;

    /// <summary>
    /// List of output records. Depending on output type, record type can be:
    /// * path to file for  AspriseImaging#OUTPUT_SAVE, AspriseImaging#OUTPUT_SAVE_THUMB;
    /// * server response for AspriseImaging#OUTPUT_UPLOAD, AspriseImaging#OUTPUT_UPLOAD_THUMB;
    /// * image data in base64 for AspriseImaging#OUTPUT_RETURN_BASE64, AspriseImaging#OUTPUT_RETURN_BASE64_THUMB.
    /// </summary>
    private List<String> outputRecords = new List<String>();

    public ResultOutputItem(XElement xe) {
        if(!xe.Attribute("type").Value.Equals("object")) {
            throw new ImagingException("Expecting object, actual: " + xe);
        }
        type = ScanUtils.xeElementValueAsString(xe, "type");
        format = ScanUtils.normalizeImageFormat(ScanUtils.xeElementValueAsString(xe, "format"));
        List<object> list = ScanUtils.xeElementArrayToList(xe.Element("result"));
        foreach(object item in list) {
            outputRecords.Add((string)item);
        }
    }

    /// <summary>
    /// Deserialize from JSON.
    /// </summary>
    /// <param name="xe">Input JSON</param>
    /// <returns>The output item deserialized or null if the argument passed is null.</returns>
    public static ResultOutputItem createScanResultOutputItem(XElement xe) {
        if(xe == null) {
            return null;
        } else {
            return new ResultOutputItem(xe);
        }
    }

    /// <summary>
    /// Output type - any of AspriseImaging#OUTPUT_SAVE ...
    /// </summary>
    /// <returns></returns>
    public String GetOutputType() {
        return type;
    }

    /// <summary>
    /// Image format - any of AspriseImaging#FORMAT_JPG ...
    /// </summary>
    /// <returns></returns>
    public String GetFormat() {
        return format;
    }


    /// <summary>
    /// Number of records.
    /// </summary>
    /// <returns></returns>
    public int GetOutputRecordCount() {
        return outputRecords == null ? 0 : outputRecords.Count;
    }


    /// <summary>
    /// List of output records. Depending on output type, record type can be:
    /// * path to file for AspriseImaging#OUTPUT_SAVE, AspriseImaging#OUTPUT_SAVE_THUMB;
    /// * server response for AspriseImaging#OUTPUT_UPLOAD, AspriseImaging#OUTPUT_UPLOAD_THUMB;
    /// * image data in base64 for AspriseImaging#OUTPUT_RETURN_BASE64, AspriseImaging#OUTPUT_RETURN_BASE64_THUMB.
    /// </summary>
    /// <returns>List of output records or empty list if there is no record.</returns>
    public List<String> GetOutputRecords() {
        return outputRecords;
    }

    /// <summary>
    /// Sets the output records.
    /// </summary>
    /// <param name="outputRecords"></param>
    public void SetOutputRecords(List<String> outputRecords) {
        this.outputRecords = outputRecords;
    }

    /// <summary>
    /// List of output files for AspriseImaging#OUTPUT_SAVE and AspriseImaging#OUTPUT_SAVE_THUMB.
    /// </summary>
    /// <returns></returns>
    public List<string> GetFiles() {
        if(!AspriseImaging.OUTPUT_SAVE.Equals(type) && !AspriseImaging.OUTPUT_SAVE_THUMB.Equals(type)) {
            return null;
        }
        List<string> files = new List<string>();
        for(int i = 0; outputRecords != null && i < outputRecords.Count; i++) {
            files.Add(outputRecords[i]);
        }
        return files;
    }

    /// <summary>
    /// Loads the image represented by the given record if output type is one of: AspriseImaging#OUTPUT_SAVE,
    /// OUTPUT_SAVE_THUMB, OUTPUT_RETURN_BASE64, OUTPUT_RETURN_BASE64_THUMB.
    /// </summary>
    /// <param name="recordIndex"></param>
    /// <returns></returns>
    public Bitmap LoadImage(int recordIndex) {
        String record = outputRecords[recordIndex];
        if(record == null) {
            return null;
        }

        Bitmap image = null;
        if(AspriseImaging.OUTPUT_SAVE.Equals(type) || AspriseImaging.OUTPUT_SAVE_THUMB.Equals(type)) {
            String file = null;
            if(AspriseImaging.FORMAT_TIF.Equals(format)) { // multi-page
                if(outputRecords.Count == 1) {
                    file = outputRecords[0];
                }
            }
            else if (AspriseImaging.FORMAT_PDF.Equals(format))
            { // PDF
                throw new ImagingException("PDF reading is not supported.");
            } else { // bmp, jpg, etc.
                file = outputRecords[recordIndex];
            }
            image = new Bitmap(file);
        } else if(AspriseImaging.OUTPUT_RETURN_BASE64.Equals(type) || AspriseImaging.OUTPUT_RETURN_BASE64_THUMB.Equals(type)) {
            String base64 = null;
            if (AspriseImaging.FORMAT_TIF.Equals(format))
            { // multi-page
                if(outputRecords.Count == 1) {
                    base64 = outputRecords[0];
                }
            }
            else if (AspriseImaging.FORMAT_PDF.Equals(format))
            { // PDF
                throw new ImagingException("PDF reading is not supported.");
            } else { // bmp, jpg, etc.
                base64 = outputRecords[recordIndex];
            }
            image = ScanUtils.newBitmapFromBase64(base64);
        } else {
            throw new ImagingException("You can load image when output type is " + type);
        }

        return image;
    }


    public override String ToString() {
        return "ScanResultOutputItem{" +
                "type='" + type + '\'' +
                ", format='" + format + '\'' +
                ", outputRecords=" + outputRecords +
                '}';
    }
}

    /// <summary>
    /// Info and metadata about the image scanned.
    /// </summary>
    public class ResultImageItem {
    /// <summary>
        /// List of barcodes recognized on the image.
    /// </summary>
    private List<Dictionary<String, Object>> barcodes = new List<Dictionary<String, Object>>();
    /// <summary>
    /// Caps returned
    /// </summary>
    private Dictionary<String, Object> caps = new Dictionary<String, Object>();
    /// <summary>
    /// Extended image attributes
    /// </summary>
    private Dictionary<String, Object> extendedImageAttrs = new Dictionary<String, Object>();
    /// <summary>
    /// Image info if available.
    /// </summary>
    private ImageInfo imageInfo;
    /// <summary>
    /// Image layout if available
    /// </summary>
    private ImageLayout imageLayout;

    private int acquiredOn;

    /// <summary>
    /// Whether it is blank
    /// </summary>
    private Boolean blank;

    /// <summary>
    /// Ink coverage
    /// </summary>
    private decimal inkCoverage = -1;

    public ResultImageItem(XElement xe) {
        object barcodeList = ScanUtils.xeElementValue(xe.Element("barcodes"));
        if (barcodeList is List<object>)
        {
            foreach(object barcodeItem in (List<object>)barcodeList) {
                barcodes.Add((Dictionary<string, object>) barcodeItem);
            }
        }
        caps = (Dictionary<String, Object>) ScanUtils.xeElementValue(xe.Element("caps"));
        extendedImageAttrs = (Dictionary<String, Object>) ScanUtils.xeElementValue(xe.Element("extended_image_attrs"));
        imageInfo = xe.Element("image_info") == null ? null : ImageInfo.createImageInfo(xe.Element("image_info"));
        imageLayout = xe.Element("image_layout") == null ? null : ImageLayout.createImageLayout(xe.Element("image_layout"));
        acquiredOn = ScanUtils.toInteger(xe.Element("time").Value, 0);
        blank = xe.Element("is_blank") == null ? false : xe.Element("is_blank").Value.Equals("true");
        inkCoverage = ScanUtils.xeElementValueAsDecimal(xe, "ink_coverage", -1);
    }

        /// <summary>
        /// Deserialize from JSON.
        /// </summary>
        /// <param name="xe"></param>
        /// <returns>The image item or null if the argument passed is null.</returns>
    public static ResultImageItem createScanResultImageItem(XElement xe) {
        if(xe == null) {
            return null;
        } else {
            return new ResultImageItem(xe);
        }
    }

    /// <summary>
    /// Caps in key(cap name)-value map.
    /// </summary>
    /// <returns></returns>
    public Dictionary<String, Object> getCaps() {
        return caps;
    }

    /// <summary>
    /// Retrieves the value of the given cap.
    /// </summary>
    /// <param name="capCode"></param>
    /// <returns>the cap value or null if cap hasn't been retrieved.</returns>
    public object GetCap(int capCode) {
        if(caps == null) {
            return null;
        }
        String capName = TwainConstants.getCapName(capCode);
        Object value = caps != null && caps.ContainsKey(capName) ? caps[capName] : null;
        return value;
    }

    /// <summary>
    /// Extended image attributes in key(extended image attribute name)-value map.
    /// </summary>
    /// <returns></returns>
    public Dictionary<String, Object> getExtendedImageAttrs() {
        return extendedImageAttrs;
    }

    /// <summary>
    /// Retrieves the value of the given attribute.
    /// </summary>
    /// <param name="attrCode"></param>
    /// <returns>the attribute value or null if it hasn't been retrieved.</returns>
    public object GetExtendedImageAttr(int attrCode) {
        if(extendedImageAttrs == null) {
            return null;
        }
        String tweiName = TwainConstants.getTweiName(attrCode);
        Object value = extendedImageAttrs != null && extendedImageAttrs.ContainsKey(tweiName) ?
            extendedImageAttrs[tweiName] : null;
        return value;
    }

    /// <summary>
    /// Image info if available.
    /// </summary>
    /// <returns></returns>
    public ImageInfo GetImageInfo() {
        return imageInfo;
    }

    /// <summary>
    /// Gets the image resolution in x direction.
    /// </summary>
    /// <returns>the image resolution or -1 if not available.</returns>
    public int GetImageInfoResolution() {
        return imageInfo == null ? -1 : imageInfo.XResolution;
    }

    /// <summary>
    /// Gets the image width.
    /// </summary>
    /// <returns>the image width or -1 if not available.</returns>
    public int GetImageInfoWidth() {
        return imageInfo == null ? -1 : imageInfo.ImageWidth;
    }

    /// <summary>
    /// Gets the image height.
    /// </summary>
    /// <returns>the image height or -1 not available.</returns>
    public int GetImageInfoHeight() {
        return imageInfo == null ? -1 : imageInfo.ImageLength;
    }

    /// <summary>
    /// Gets pixel type.
    /// </summary>
    /// <returns>-1 if not available or any of TwainConstants#TWPT_BW, TwainConstants#TWPT_GRAY or TwainConstants#TWPT_RGB.</returns>
    public int GetImageInfoPixelType() {
        return imageInfo == null ? -1 : imageInfo.PixelType;
    }

    /// <summary>
    /// Gets the bit depth (bits per pixel).
    /// </summary>
    /// <returns>-1 if not available or the actual bit depth.</returns>
    public int GetImageInfoBitDepth() {
        return imageInfo == null ? -1 : imageInfo.BitsPerPixel;
    }

    /// <summary>
    /// Image layout if available.
    /// </summary>
    /// <returns></returns>
    public ImageLayout GetImageLayout() {
        return imageLayout;
    }

    /// <summary>
    /// Gets the document number.
    /// </summary>
    /// <returns>the document number or -1 if not available.</returns>
    public int GetImageLayoutDocumentNumber() {
        return imageLayout == null ? -1 : imageLayout.DocumentNumber;
    }

    /// <summary>
    /// Gets the frame number.
    /// </summary>
    /// <returns>the frame number or -1 if not available.</returns>
    public int GetImageLayoutFrameNumber() {
        return imageLayout == null ? -1 : imageLayout.FrameNumber;
    }

    /// <summary>
    /// Gets the page number.
    /// </summary>
    /// <returns>the page number or -1 if not available.</returns>
    public int GetImageLayoutPageNumber() {
        return imageLayout == null ? -1 : imageLayout.PageNumber;
    }

    /// <summary>
    /// Gets the frame location (left, top, right, bottom) in unit TwainConstants#ICAP_UNITS.
    /// </summary>
    /// <returns>the frame location (4-element double array) or null if not available.</returns>
    public decimal[] GetImageLayoutFrame() {
        return imageLayout == null ? null : imageLayout.Frame;
    }

    /// <summary>
    /// The acquisition time in epoch second or 0 if not available.
    /// </summary>
    /// <returns></returns>
    public int GetAcquiredOn() {
        return acquiredOn;
    }

    /// <summary>
    /// List of barcodes recognized on the image.
    /// </summary>
    /// <returns>barcodes recognized or empty list if recognition is not performed or none is recognized.</returns>
    public List<Dictionary<String, Object>> GetBarcodes() {
        return barcodes;
    }

    /// <summary>
    /// Whether it is blank - available only if blank page detection is enabled.
    /// </summary>
    /// <returns></returns>
    public Boolean IsBlank() {
        return blank;
    }

    /// <summary>
    /// Ink coverage - available only if blank page detection is enabled.
    /// </summary>
    /// <returns>ink coverage or negative value if it is not available.</returns>
    public decimal GetInkCoverage() {
        return inkCoverage;
    }

    /// <summary>
    /// Corresponding to TWAIN's TW_IMAGEINFO struct
    /// </summary>
    public class ImageInfo {
        /// <summary>
        /// Impacted by cap: ICAP_XRESOLUTION
        /// </summary>
        public int XResolution = -1;
        /// <summary>
        /// Impacted by cap: ICAP_YRESOLUTION
        /// </summary>
        public int YResolution = -1;
        /// <summary>
        /// Impacted by cap: TW_IMAGELAYOUT.TW_FRAME.Right - TW_FRAME.Left
        /// </summary>
        public int ImageWidth = -1;
        /// <summary>
        /// Impacted by cap: TW_IMAGELAYOUT.TW_FRAME.Bottom - TW_FRAME.Top
        /// </summary>
        public int ImageLength = -1;
        /// <summary>
        /// Impacted by cap: ICAP_PIXELTYPE (i.e. TWPT_BW has 1, TWPT_RGB has 3)
        /// </summary>
        public int SamplesPerPixel = -1;
        /// <summary>
        /// Calculated by BitsPerPixel divided by SamplesPerPixel.
        /// </summary>
        public int[] BitsPerSample;
        /// <summary>
        /// Impacted by cap: ICAP_BITDEPTH
        /// </summary>
        public int BitsPerPixel = -1;
        /// <summary>
        /// Impacted by cap: ICAP_PLANARCHUNKY
        /// </summary>
        public Boolean Planar = false;
        /// <summary>
        /// Impacted by cap: ICAP_PIXELTYPE
        /// </summary>
        public int PixelType = -1;
        /// <summary>
        /// Impacted by cap: ICAP_COMPRESSION
        /// </summary>
        public int Compression = -1;

        public ImageInfo(XElement xe) {
            if(! xe.Attribute("type").Value.Equals("object")) {
                throw new ImagingException("Expecting type=object, actual: " + xe);
            }
            this.XResolution = ScanUtils.xeElementValueAsInt(xe, "XResolution", -1);
            this.YResolution = ScanUtils.xeElementValueAsInt(xe, "YResolution", -1);
            this.ImageWidth = ScanUtils.xeElementValueAsInt(xe, "ImageWidth", -1);
            this.ImageLength = ScanUtils.xeElementValueAsInt(xe, "ImageLength", -1);
            this.SamplesPerPixel = ScanUtils.xeElementValueAsInt(xe, "SamplesPerPixel", -1);
            this.BitsPerPixel = ScanUtils.xeElementValueAsInt(xe, "BitsPerPixel", -1);
            this.Planar = ScanUtils.xeElementValueAsInt(xe, "Planar", 0) != 0;
            this.PixelType = ScanUtils.xeElementValueAsInt(xe, "PixelType", -1);
            this.Compression = ScanUtils.xeElementValueAsInt(xe, "Compression", -1);
        }

        public ImageInfo() {
        }

        /// <summary>
        /// Deserialize from JSON.
        /// </summary>
        /// <param name="xe"></param>
        /// <returns>Image info or null if the argument passed is null.</returns>
        public static ImageInfo createImageInfo(XElement xe) {
            if(xe == null) {
                return null;
            } else {
                return new ImageInfo(xe);
            }
        }

        public override String ToString() {
            return "ImageInfo{" +
                    "XResolution=" + XResolution +
                    ", YResolution=" + YResolution +
                    ", ImageWidth=" + ImageWidth +
                    ", ImageLength=" + ImageLength +
                    ", SamplesPerPixel=" + SamplesPerPixel +
                    ", BitsPerSample=" + (BitsPerSample == null ? "(null)" : String.Join(",", BitsPerSample)) +
                    ", BitsPerPixel=" + BitsPerPixel +
                    ", Planar=" + Planar +
                    ", PixelType=" + PixelType +
                    ", Compression=" + Compression +
                    '}';
        }
    }
    }
    
    /// <summary>
    /// Image layout info such as frame, doc number and page number.
    /// </summary>
    public class ImageLayout {
        public decimal[] Frame;
        public int DocumentNumber = -1;
        public int PageNumber = -1;
        public int FrameNumber = -1;

        public ImageLayout(XElement xe) {
            if(! xe.Attribute("type").Value.Equals("object")) {
                throw new ImagingException("Excepted type=object, actual: " + xe);
            }
            this.DocumentNumber = ScanUtils.xeElementValueAsInt(xe, "DocumentNumber", -1);
            this.PageNumber = ScanUtils.xeElementValueAsInt(xe, "PageNumber", -1);
            this.FrameNumber = ScanUtils.xeElementValueAsInt(xe, "FrameNumber", -1);
            this.Frame = ScanUtils.parseFrame(xe.Element("Frame").Value);
        }

        public ImageLayout() {
        }

        public static ImageLayout createImageLayout(XElement xe) {
            if(xe == null) {
                return null;
            } else {
                return new ImageLayout(xe);
            }
        }

        public override String ToString() {
            return "ImageLayout{" +
                    "Frame=" + String.Join("|", Frame) +
                    ", DocumentNumber=" + DocumentNumber +
                    ", PageNumber=" + PageNumber +
                    ", FrameNumber=" + FrameNumber +
                    '}';
        }
    }
}
